"""dataflow package provides a framework to build data analysis pipeline

class:
    Flow -- main class for the dataflow package ;
        provides functionnal tools to transform a dataset
        and do machine learning with it
"""

from flow import Flow, Action

__all__ = [Flow, Action]

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



# -*- coding: utf-8 -*-

from flow import Flow


class TestFlow:

    def test_init(self):
        """Test initialization"""

        flow = Flow()
        assert hasattr(flow, 'data')
        assert hasattr(flow, 'source')
        assert hasattr(flow, 'chain')
        assert flow.data is None

    def test_filter(self, tmpdir_factory):
        """Test filter registering"""

        p = tmpdir_factory.mktemp("data").join("test.csv")

        p.write('\n'.join(
            [','.join(['first', 'line'])] + [
                ','.join([str(i) for i in range(5)]) for _ in range(10)
            ] + [','.join(['last', 'line'])]))

        flow = Flow().from_file(p.open()).filter(
            lambda x: 'line' not in x
        )

        assert flow.chain[0].type == 'FLOW::FILTER'
        assert flow.batch(1)[0] == ['0', '1', '2', '3', '4']

    def test_map(self, tmpdir_factory):
        """Test map registering"""

        p = tmpdir_factory.mktemp("data").join("test.csv")

        p.write('\n'.join(
            [','.join([str(i) for i in range(5)]) for _ in range(10)]
        ))

        flow = Flow().from_file(p.open()).map(
            lambda x: [float(_) for _ in x]
        ).map(
            lambda x: [2 * _ for _ in x]
        )

        assert flow.chain[0].type == 'FLOW::MAP'
        assert flow.batch(1)[0] == [0, 2, 4, 6, 8]

    def test_reduce(self, tmpdir_factory):
        """Test reduce"""

        p = tmpdir_factory.mktemp("data").join("test.csv")

        p.write('\n'.join(
            [','.join([str(i) for i in range(5)]) for _ in range(10)]
        ))

        flow = Flow().from_file(p.open()).map(
            lambda x: [float(_) for _ in x]
        ).map(lambda x: sum(x))

        assert flow.reduce(lambda a, b: a + b) == 100

    def test_reload(self, tmpdir_factory):
        """Test reloading"""

        p = tmpdir_factory.mktemp("data").join("test.csv")

        p.write('\n'.join(
            [','.join([str(i) for i in range(5)]) for _ in range(10)]
        ))

        flow = Flow().from_file(p.open()).map(lambda x: 1)

        assert flow.reduce(lambda a, b: a + b) == 10
        assert flow.reload().reduce(lambda a, b: a + b) == 10



# -*- coding: utf-8 -*-
from collections import namedtuple
import itertools


"""Action namedtuple used to differentiate map transformation from filter op"""
Action = namedtuple('action', ['type', 'f'])


def read_csv(path, sep, line_terminator):
    """Yield a value from a csv file on disk

    Keyword arguments:
    path            (string) -- path to file
    sep             (string) -- separator substring
    line_terminator (string) -- line terminator substring
    """

    with open(path) as fin:

        for line in fin:
            yield line.replace('"', '').replace(line_terminator, '').split(sep)


def read_file(fin, sep, line_terminator):
    """Yield a value from a csv file in memory

    Keyword arguments:
    path            (string) -- path to file
    sep             (string) -- separator substring
    line_terminator (string) -- line terminator substring
    """

    fin.seek(0)

    for line in fin:
        yield line.replace('"', '').replace(line_terminator, '').split(sep)


def apply_action(data, action):
    """Compose successive generators from ordered map and filter transformation
    """

    if action.type == 'FLOW::MAP':
        return (action.f(_) for _ in data)

    elif action.type == 'FLOW::FILTER':
        return (_ for _ in data if action.f(_))

    else:
        return data


class Flow(object):

    def __init__(self):
        """Init Flow object"""

        self.data = None
        self.source = None
        self.chain = []

    def from_csv(self, path, sep=',', line_terminator='\n'):
        """Hook Flow to disk csv source

        Keyword arguments:
        path            (string) -- path to file
        sep             (string) -- separator substring
        line_terminator (string) -- line terminator substring
        """

        self.data = read_csv(path, sep, line_terminator)
        self.source = (read_csv, path, sep, line_terminator)

        return self

    def from_file(self, fin, sep=',', line_terminator='\n'):
        """Hook Flow to memory csv source

        Keyword arguments:
        path            (string) -- path to file
        sep             (string) -- separator substring
        line_terminator (string) -- line terminator substring
        """

        self.data = read_file(fin, sep, line_terminator)
        self.source = (read_file, fin, sep, line_terminator)

        return self

    def map(self, f):
        """Register a map transformation of function f

        Keyword arguments:
        f               (function) -- function with signature x => value
        """

        self.chain += [Action('FLOW::MAP', f)]

        return self

    def filter(self, f):
        """Register a filter transformation of function f

        Keyword arguments:
        f               (function) -- function with signature x => boolean
        """

        self.chain += [Action('FLOW::FILTER', f)]

        return self

    def reduce(self, f):
        """Reduce data with function f

        Keyword arguments:
        f               (function) -- function with signature a, b => value
        """

        return reduce(f, self.flow().data)

    def eval(self):
        """Evaluate dataset

        Apply all successive transformations and return resulting dataset
        """

        return [_ for _ in self.flow().data]

    def batch(self, size):
        """Evaluate dataset

        Apply all successive transformations and return resulting dataset slice

        Keyword arguments:
        size            (int) -- size of slice
        """

        return list(itertools.islice(self.flow().data, size))

    def flow(self):
        """Compose generator from successive registered actions"""

        for a in self.chain:
            self.data = apply_action(self.data, a)

        return self

    def reload(self):
        """Renew the self.data generator"""

        self.data = self.source[0](*self.source[1:])

        return self

    def __repr__(self):
        """Evaluate whole dataset and return string representation"""

        return '\n'.join(self.eval()) if self.data is not None else 'None'



# -*- coding: utf-8 -*-

from distutils.core import setup

import versioneer

setup(
    name='waterflow',
    packages=['waterflow'],
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Dataflow package provides a data analysis pipeline' +
    'framework for data transformation and machine learning',
    author='Aurélien Demilly',
    author_email='demilly.aurelien@gmail.com',
    url="https://github.com/ademilly/dataflow",
    keywords=['data science', 'data analysis', 'machine learning'],
    classifiers=[],
)
